# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gpyrpc/protorpc_wire.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1agpyrpc/protorpc_wire.proto\x12\rprotorpc_wire"}\n\rRequestHeader\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x17\n\x0fraw_request_len\x18\x03 \x01(\r\x12%\n\x1dsnappy_compressed_request_len\x18\x04 \x01(\r\x12\x10\n\x08\x63hecksum\x18\x05 \x01(\r"\x7f\n\x0eResponseHeader\x12\n\n\x02id\x18\x01 \x01(\x04\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x18\n\x10raw_response_len\x18\x03 \x01(\r\x12&\n\x1esnappy_compressed_response_len\x18\x04 \x01(\r\x12\x10\n\x08\x63hecksum\x18\x05 \x01(\r*.\n\x05\x43onst\x12\x08\n\x04ZERO\x10\x00\x12\x1b\n\x16MAX_REQUEST_HEADER_LEN\x10\x80\x08\x62\x06proto3'
)

_CONST = DESCRIPTOR.enum_types_by_name["Const"]
Const = enum_type_wrapper.EnumTypeWrapper(_CONST)
ZERO = 0
MAX_REQUEST_HEADER_LEN = 1024


_REQUESTHEADER = DESCRIPTOR.message_types_by_name["RequestHeader"]
_RESPONSEHEADER = DESCRIPTOR.message_types_by_name["ResponseHeader"]
RequestHeader = _reflection.GeneratedProtocolMessageType(
    "RequestHeader",
    (_message.Message,),
    {
        "DESCRIPTOR": _REQUESTHEADER,
        "__module__": "gpyrpc.protorpc_wire_pb2"
        # @@protoc_insertion_point(class_scope:protorpc_wire.RequestHeader)
    },
)
_sym_db.RegisterMessage(RequestHeader)

ResponseHeader = _reflection.GeneratedProtocolMessageType(
    "ResponseHeader",
    (_message.Message,),
    {
        "DESCRIPTOR": _RESPONSEHEADER,
        "__module__": "gpyrpc.protorpc_wire_pb2"
        # @@protoc_insertion_point(class_scope:protorpc_wire.ResponseHeader)
    },
)
_sym_db.RegisterMessage(ResponseHeader)

if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _CONST._serialized_start = 301
    _CONST._serialized_end = 347
    _REQUESTHEADER._serialized_start = 45
    _REQUESTHEADER._serialized_end = 170
    _RESPONSEHEADER._serialized_start = 172
    _RESPONSEHEADER._serialized_end = 299
# @@protoc_insertion_point(module_scope)
