# Copyright 2021 The KCL Authors. All rights reserved.


class TokenValue:
    SEMI = ";"
    COMMA = ","
    EQ = "="
    LPAREN = "("
    RPAREN = ")"
    LBRACK = "["
    RBRACK = "]"
    LBRACE = "{"
    RBRACE = "}"
    LANGLE_BRACK = "<"
    RANGLE_BRACK = ">"
    SYNTAX = "syntax"
    IMPORT = "import"
    MESSAGE = "message"
    ENUM = "enum"
    PACKAGE = "package"
    OPTION = "option"
    ONEOF = "oneof"
    MAP = "map"
    REPEATED = "repeated"
    SERVICE = "service"
    RPC = "rpc"
    RETURNS = "returns"
    TO = "to"
    RESERVED = "reserved"
