"""
Base id of standard checkers (used in msg and report ids):
01: basic
02: Schema
03: format
04: import
05: misc
06: variables
07:
08：
09: syntax
10-50: not yet used: reserved for future internal checkers.
51-99: perhaps used: reserved for external checkers
"""

__all__ = ["base_checker", "imports", "misc", "basic"]
